@@ -5,13 +5,9 @@ from decimal import Decimal
 from itertools import groupby
 from functools import partial
 
-from sql import Null
-from sql.conditionals import Case
-
-from trytond import backend
 from trytond.model import ModelView, ModelSQL, fields
 from trytond.wizard import Wizard, StateView, StateTransition, Button
-from trytond.pyson import If, In, Eval
+from trytond.pyson import If, In, Eval, Bool
 from trytond.transaction import Transaction
 from trytond.pool import Pool
 
@@ -31,14 +27,20 @@ class PurchaseRequest(ModelSQL, ModelView):
     'Purchase Request'
     __name__ = 'purchase.request'
 
-    product = fields.Many2One('product.product', 'Product', required=True,
+    product = fields.Many2One('product.product', 'Product',
         select=True, readonly=True, domain=[('purchasable', '=', True)])
+    description = fields.Text('Description', readonly=True,
+        states=STATES, depends=DEPENDS)
     party = fields.Many2One('party.party', 'Party', select=True, states=STATES,
         depends=DEPENDS)
     quantity = fields.Float('Quantity', required=True, states=STATES,
         digits=(16, Eval('uom_digits', 2)), depends=DEPENDS + ['uom_digits'])
-    uom = fields.Many2One('product.uom', 'UOM', required=True, select=True,
-        states=STATES, depends=DEPENDS)
+    uom = fields.Many2One('product.uom', 'UOM', select=True,
+        states={
+            'required': Bool(Eval('product')),
+            'readonly': STATES['readonly'],
+            },
+        depends=['product'] + DEPENDS)
     uom_digits = fields.Function(fields.Integer('UOM Digits'),
         'on_change_with_uom_digits')
     computed_quantity = fields.Float('Computed Quantity', readonly=True)
@@ -71,13 +73,13 @@ class PurchaseRequest(ModelSQL, ModelView):
             ])
     origin = fields.Reference('Origin', selection='get_origin', readonly=True)
     exception_ignored = fields.Boolean('Ignored Exception')
-    state = fields.Function(fields.Selection([
-        ('purchased', 'Purchased'),
-        ('done', 'Done'),
-        ('draft', 'Draft'),
-        ('cancel', 'Cancel'),
-        ('exception', 'Exception'),
-        ], 'State'), 'get_state', searcher='search_state')
+    state = fields.Selection([
+            ('purchased', "Purchased"),
+            ('done', "Done"),
+            ('draft', "Draft"),
+            ('cancel', "Cancel"),
+            ('exception', "Exception"),
+            ], "State", required=True, readonly=True, select=True)
 
     @classmethod
     def __setup__(cls):
@@ -92,6 +94,7 @@ class PurchaseRequest(ModelSQL, ModelView):
         cls._buttons.update({
                 'handle_purchase_cancellation_exception': {
                     'invisible': Eval('state') != 'exception',
+                    'depends': ['state'],
                     },
                 })
 
@@ -99,12 +102,20 @@ class PurchaseRequest(ModelSQL, ModelView):
     def __register__(cls, module_name):
         pool = Pool()
         ModelData = pool.get('ir.model.data')
-        TableHandler = backend.get('TableHandler')
+        Purchase = pool.get('purchase.purchase')
+        PurchaseLine = pool.get('purchase.line')
         model_data = ModelData.__table__()
+        purchase = Purchase.__table__()
+        purchase_line = PurchaseLine.__table__()
+        request = cls.__table__()
+
+        tablehandler = cls.__table_handler__(module_name)
+        state_exist = tablehandler.column_exist('state')
+
         super(PurchaseRequest, cls).__register__(module_name)
 
         # Migration from 3.6: removing the constraint on the quantity
-        tablehandler = TableHandler(cls, module_name)
+        tablehandler = cls.__table_handler__(module_name)
         tablehandler.drop_constraint('check_purchase_request_quantity')
 
         # Migration from 3.8: renaming module of Purchase Request group entry
@@ -115,20 +126,69 @@ class PurchaseRequest(ModelSQL, ModelView):
                 where=((model_data.fs_id == 'group_purchase_request')
                     & (model_data.module == 'stock_supply'))))
 
+        # Migration from 4.0: remove required on product and uom
+        tablehandler.not_null_action('product', action='remove')
+        tablehandler.not_null_action('uom', action='remove')
+
+        # Migration from 4.2: add state
+        if not state_exist:
+            cursor = Transaction().connection.cursor()
+            update = Transaction().connection.cursor()
+            query = request.join(purchase_line, type_='INNER',
+                condition=request.purchase_line == purchase_line.id
+                ).join(purchase, type_='INNER',
+                    condition=purchase_line.purchase == purchase.id
+                    ).select(
+                        request.id, purchase.state, request.exception_ignored)
+            cursor.execute(*query)
+            for request_id, purchase_state, exception_ignored in cursor:
+                if purchase_state == 'cancel' and not exception_ignored:
+                    state = 'exception'
+                elif purchase_state == 'cancel':
+                    state = 'cancel'
+                elif purchase_state == 'done':
+                    state = 'done'
+                else:
+                    state = 'purchased'
+                update.execute(*request.update(
+                        [request.state],
+                        [state],
+                        where=request.id == request_id))
+
+        # Migration from 4.4: remove required on origin
+        tablehandler.not_null_action('origin', action='remove')
+
     def get_rec_name(self, name):
-        if self.warehouse:
-            return "%s@%s" % (self.product.name, self.warehouse.name)
+        if self.product:
+            rec_name = self.product.name
+        elif self.description:
+            rec_name = self.description.splitlines()[0]
         else:
-            return self.product.name
+            rec_name = str(self.id)
+
+        if self.warehouse:
+            return "%s@%s" % (rec_name, self.warehouse.name)
+        else:
+            return rec_name
 
     @classmethod
     def search_rec_name(cls, name, clause):
-        res = []
-        names = clause[2].split('@', 1)
-        res.append(('product.template.name', clause[1], names[0]))
-        if len(names) != 1 and names[1]:
-            res.append(('warehouse', clause[1], names[1]))
-        return res
+        _, operator, value = clause
+        if operator.startswith('!') or operator.startswith('not'):
+            bool_op = 'AND'
+        else:
+            bool_op = 'OR'
+        domain = [bool_op]
+        if value is not None:
+            names = value.split('@', 1)
+            sub_domain = [('product.template.name', operator, names[0])]
+            if len(names) != 1 and names[1]:
+                sub_domain.append(('warehouse', operator, names[1]))
+                if bool_op == 'AND':
+                    sub_domain.insert(0, 'OR')
+            domain.append(sub_domain)
+        domain.append(('description', *clause[1:]))
+        return domain
 
     @staticmethod
     def default_company():
@@ -150,7 +210,11 @@ class PurchaseRequest(ModelSQL, ModelView):
     def currency(self):
         return self.company.currency
 
-    def get_state(self, name):
+    @classmethod
+    def default_state(cls):
+        return 'draft'
+
+    def get_state(self):
         if self.purchase_line:
             if (self.purchase_line.purchase.state == 'cancel'
                     and not self.exception_ignored):
@@ -164,38 +228,15 @@ class PurchaseRequest(ModelSQL, ModelView):
         return 'draft'
 
     @classmethod
-    def search_state(cls, name, clause):
-        pool = Pool()
-        Purchase = pool.get('purchase.purchase')
-        PurchaseLine = pool.get('purchase.line')
-
-        request = cls.__table__()
-        purchase_line = PurchaseLine.__table__()
-        purchase = Purchase.__table__()
-
-        _, operator_, state = clause
-        Operator = fields.SQL_OPERATORS[operator_]
-        state_case = Case(
-            ((purchase.state == 'cancel')
-                & (request.exception_ignored == False), 'exception'),
-            ((purchase.state == 'cancel')
-                & (request.exception_ignored == True), 'cancel'),
-            (purchase.state == 'done', 'done'),
-            (request.purchase_line != Null, 'purchased'),
-            else_='draft')
-        state_query = request.join(
-            purchase_line, type_='LEFT',
-            condition=request.purchase_line == purchase_line.id
-            ).join(purchase, type_='LEFT',
-            condition=purchase_line.purchase == purchase.id
-            ).select(
-            request.id,
-            where=Operator(state_case, state))
-
-        return [('id', 'in', state_query)]
+    def update_state(cls, requests):
+        for request in requests:
+            state = request.get_state()
+            if state != request.state:
+                request.state = state
+        cls.save(requests)
 
     def get_warehouse_required(self, name):
-        return self.product.type in ('goods', 'assets')
+        return self.product and self.product.type in ('goods', 'assets')
 
     @fields.depends('uom')
     def on_change_with_uom_digits(self, name=None):
@@ -221,12 +262,12 @@ class PurchaseRequest(ModelSQL, ModelView):
         models = IrModel.search([
                 ('model', 'in', list(cls._get_origin())),
                 ])
-        return [(m.model, m.name) for m in models]
+        return [(None, '')] + [(m.model, m.name) for m in models]
 
     @classmethod
     def create(cls, vlist):
         for vals in vlist:
-            for field_name in ('product', 'quantity', 'uom', 'company'):
+            for field_name in ('quantity', 'company'):
                 if vals.get(field_name) is None:
                     cls.raise_user_error('create_request')
         return super(PurchaseRequest, cls).create(vlist)
@@ -270,6 +311,7 @@ class CreatePurchaseAskParty(ModelView):
     'Create Purchase Ask Party'
     __name__ = 'purchase.request.create_purchase.ask_party'
     product = fields.Many2One('product.product', 'Product', readonly=True)
+    description = fields.Text('Description', readonly=True)
     company = fields.Many2One('company.company', 'Company', readonly=True)
     party = fields.Many2One('party.party', 'Supplier', required=True)
 
@@ -281,7 +323,7 @@ class CreatePurchase(Wizard):
     ask_party = StateView('purchase.request.create_purchase.ask_party',
         'purchase_request.purchase_request_create_purchase_ask_party', [
             Button('Cancel', 'end', 'tryton-cancel'),
-            Button('Continue', 'start', 'tryton-go-next', default=True),
+            Button('Continue', 'start', 'tryton-forward', default=True),
             ])
 
     @classmethod
@@ -300,12 +342,10 @@ class CreatePurchase(Wizard):
             if request.purchase_line:
                 continue
             if not request.party:
-                return {
-                    'product': request.product.id,
-                    'company': request.company.id,
-                    }
+                break
         return {
-            'product': request.product.id,
+            'product': request.product.id if request.product else None,
+            'description': request.description,
             'company': request.company.id,
             }
 
@@ -324,45 +364,62 @@ class CreatePurchase(Wizard):
             ('invoice_address', request.party.address_get(type='invoice')),
             )
 
+    def _group_purchase_line_key(self, request):
+        '''
+        The key to group requests by lines
+        A list of key-value as tuples of the purchase line
+        '''
+        return (
+            ('product', request.product),
+            ('description', request.description),
+            ('unit', request.uom),
+            )
+
     def transition_start(self):
         pool = Pool()
         Request = pool.get('purchase.request')
         Purchase = pool.get('purchase.purchase')
+        Line = pool.get('purchase.line')
         Date = pool.get('ir.date')
 
-        request_ids = Transaction().context['active_ids']
+        requests = Request.browse(Transaction().context['active_ids'])
 
-        if (getattr(self.ask_party, 'product', None)
-                and getattr(self.ask_party, 'party', None)
+        if (getattr(self.ask_party, 'party', None)
                 and getattr(self.ask_party, 'company', None)):
-            reqs = Request.search([
-                    ('id', 'in', request_ids),
-                    ('party', '=', None),
-                    ])
+            def compare_string(first, second):
+                return (first or '') == (second or '')
+
+            def to_write(request):
+                return (not request.purchase_line
+                    and not request.party
+                    and request.product == self.ask_party.product
+                    and compare_string(
+                        request.description, self.ask_party.description))
+            reqs = list(filter(to_write, requests))
             if reqs:
                 Request.write(reqs, {
                         'party': self.ask_party.party.id,
                         })
             self.ask_party.product = None
+            self.ask_party.description = None
             self.ask_party.party = None
             self.ask_party.company = None
 
-        reqs = Request.search([
-                ('id', 'in', request_ids),
-                ('purchase_line', '=', None),
-                ('party', '=', None),
-                ])
-        if reqs:
+        def to_ask_party(request):
+            return not request.purchase_line and not request.party
+        reqs = filter(to_ask_party, requests)
+        if any(reqs):
             return 'ask_party'
 
         today = Date.today()
-        requests = Request.browse(request_ids)
 
         requests = [r for r in requests if not r.purchase_line]
 
         keyfunc = partial(self._group_purchase_key, requests)
         requests = sorted(requests, key=keyfunc)
 
+        purchases = []
+        lines = []
         for key, grouped_requests in groupby(requests, key=keyfunc):
             grouped_requests = list(grouped_requests)
             try:
@@ -376,60 +433,33 @@ class CreatePurchase(Wizard):
             purchase = Purchase(purchase_date=purchase_date)
             for f, v in key:
                 setattr(purchase, f, v)
-            purchase.save()
-            for request in grouped_requests:
-                line = self.compute_purchase_line(request, purchase)
+            purchases.append(purchase)
+            for line_key, line_requests in groupby(
+                    grouped_requests, key=self._group_purchase_line_key):
+                line_requests = list(line_requests)
+                line = self.compute_purchase_line(
+                    line_key, line_requests, purchase)
                 line.purchase = purchase
-                line.save()
-                Request.write([request], {
-                        'purchase_line': line.id,
-                        })
+                line.requests = line_requests
+                lines.append(line)
+        Purchase.save(purchases)
+        Line.save(lines)
+        Request.update_state(requests)
         return 'end'
 
-    @staticmethod
-    def _get_tax_rule_pattern(request):
-        '''
-        Get tax rule pattern
-        '''
-        return {}
-
     @classmethod
-    def compute_purchase_line(cls, request, purchase):
+    def compute_purchase_line(cls, key, requests, purchase):
         pool = Pool()
-        Product = pool.get('product.product')
         Line = pool.get('purchase.line')
 
-        line = Line(
-            product=request.product,
-            unit=request.uom,
-            quantity=request.quantity,
-            description=request.product.name,
-            )
-
-        # XXX purchase with several lines of the same product
-        with Transaction().set_context(uom=request.uom.id,
-                supplier=request.party.id,
-                currency=request.currency.id):
-            product_price = Product.get_purchase_price(
-                [request.product], request.quantity)[request.product.id]
-            product_price = product_price.quantize(
-                Decimal(1) / 10 ** Line.unit_price.digits[1])
-
-        if product_price is None:
-            cls.raise_user_error('missing_price', (request.product.rec_name,),
-                'please_update')
-        line.unit_price = product_price
-
-        taxes = []
-        for tax in request.product.supplier_taxes_used:
-            if request.party and request.party.supplier_tax_rule:
-                pattern = cls._get_tax_rule_pattern(request)
-                tax_ids = request.party.supplier_tax_rule.apply(tax, pattern)
-                if tax_ids:
-                    taxes.extend(tax_ids)
-                continue
-            taxes.append(tax.id)
-        line.taxes = taxes
+        line = Line()
+        line.unit_price = Decimal(0).quantize(
+            Decimal(1) / 10 ** Line.unit_price.digits[1])
+        for f, v in key:
+            setattr(line, f, v)
+        line.quantity = sum(r.quantity for r in requests)
+        line.purchase = purchase
+        line.on_change_product()
         return line
 
 
@@ -452,9 +482,9 @@ class HandlePurchaseCancellationException(Wizard):
         PurchaseRequest = pool.get('purchase.request')
 
         requests = PurchaseRequest.browse(Transaction().context['active_ids'])
-        PurchaseRequest.write(requests, {
-                'purchase_line': None,
-                })
+        for request in requests:
+            request.purchase_line = None
+        PurchaseRequest.update_state(requests)
         return 'end'
 
     def transition_cancel_request(self):
@@ -462,9 +492,9 @@ class HandlePurchaseCancellationException(Wizard):
         PurchaseRequest = pool.get('purchase.request')
 
         requests = PurchaseRequest.browse(Transaction().context['active_ids'])
-        PurchaseRequest.write(requests, {
-                'exception_ignored': True,
-                })
+        for request in requests:
+            request.exception_ignored = True
+        PurchaseRequest.update_state(requests)
         return 'end'
 
 
